import logging
import os
import uuid

import pytest

from hologres_vector import HologresVector

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def get_connection_string():
    host = os.environ["HOLO_HOST"]
    port = os.environ["HOLO_PORT"]
    dbname = os.environ["HOLO_DBNAME"]
    user = os.environ["HOLO_USER"]
    password = os.environ["HOLO_PASSWORD"]

    return HologresVector.connection_string_from_db_params(host, port, dbname, user, password)


def test_connection():
    table_name = "test_vector_table_" + uuid.uuid1().hex

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema={"t": "text", "date": "timestamptz", "i": "int"},
        pre_delete_table=True,
        logger=logger,
    )

    res = holo.run_sql(
        f"select count(1) from hologres.hg_table_properties "
        f"where table_name = '{table_name}' and property_key = 'table_id'"
    )
    assert res[0][0] == 1
    holo.run_sql(f"drop table {table_name}", False)


def insert_example_data(holo: HologresVector, use_metadata: bool = False, use_id: bool = False):
    vectors = [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]]
    schema_datas = [
        {"t": "text 0", "date": "2023-08-02 13:00:00", "i": 0},
        {"t": "text 1", "date": "2023-08-02 14:00:00", "i": 1},
        {"t": "text 2", "date": "2023-08-02 15:00:00", "i": 2},
    ]
    metadatas = None
    if use_metadata:
        metadatas = schema_datas

    ids = None
    if use_id:
        ids = [0, 1, 2]

    holo.upsert_vectors(vectors, ids=ids, metadatas=metadatas, schema_datas=schema_datas)


def test_insert():
    table_name = "test_vector_table_" + uuid.uuid1().hex

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema={"t": "text", "date": "timestamptz", "i": "int"},
        pre_delete_table=True,
        logger=logger,
    )

    insert_example_data(holo)

    res = holo.run_sql(f"select vector, t, date, i, metadata from {table_name} order by i")
    assert len(res) == 3
    for i in range(3):
        assert res[i][1] == f"text {i}"
        assert res[i][3] == i
        assert res[i][4] == {}

    holo.run_sql(f"drop table {table_name}", False)


def test_upsert():
    table_name = "test_vector_table_" + uuid.uuid1().hex

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema={"t": "text", "date": "timestamptz", "i": "int"},
        pre_delete_table=True,
        logger=logger,
    )

    insert_example_data(holo, use_metadata=False, use_id=True)

    res = holo.run_sql(f"select vector, t, date, i, metadata from {table_name} order by i")
    assert len(res) == 3
    for i in range(3):
        assert res[i][1] == f"text {i}"
        assert res[i][3] == i
        assert res[i][4] == {}

    # insert again with metadata
    insert_example_data(holo, use_metadata=True, use_id=True)
    res = holo.run_sql(f"select vector, t, date, i, metadata from {table_name} order by i")
    assert len(res) == 3
    for i in range(3):
        assert res[i][1] == f"text {i}"
        assert res[i][3] == i
        assert res[i][4]["i"] == i

    holo.run_sql(f"drop table {table_name}", False)


def test_search():
    table_name = "test_vector_table_" + uuid.uuid1().hex

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema={"t": "text", "date": "timestamptz", "i": "int"},
        pre_delete_table=True,
        logger=logger,
    )
    insert_example_data(holo)

    res = holo.search([2, 2, 2, 2, 2], k=2, select_columns=["t", "date", "i"])
    assert len(res) == 2
    for i in range(2):
        assert res[i]["i"] == i
        assert res[i]["t"] == f"text {i}"

    holo.run_sql(f"drop table {table_name}", False)


def test_reconnect():
    table_name = "test_vector_table_" + uuid.uuid1().hex
    table_schema = {"t": "text", "date": "timestamptz", "i": "int"}

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema=table_schema,
        pre_delete_table=True,
        logger=logger,
    )
    insert_example_data(holo)

    holo = None  # close connection

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema=table_schema,
        logger=logger,
    )

    res = holo.search([2, 2, 2, 2, 2], k=2, select_columns=["t", "date", "i"])
    assert len(res) == 2
    for i in range(2):
        assert res[i]["i"] == i
        assert res[i]["t"] == f"text {i}"

    holo.run_sql(f"drop table {table_name}", False)


def test_metadata_filter():
    table_name = "test_vector_table_" + uuid.uuid1().hex
    table_schema = {"t": "text", "date": "timestamptz", "i": "int"}

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema=table_schema,
        pre_delete_table=True,
        logger=logger,
    )
    insert_example_data(holo, use_metadata=True)

    res = holo.search([2, 2, 2, 2, 2], k=2, select_columns=["i"], metadata_filters={"t": "text 1"})
    assert len(res) == 1
    assert res[0]["i"] == 1

    holo.run_sql(f"drop table {table_name}", False)


def test_schema_data_filter():
    table_name = "test_vector_table_" + uuid.uuid1().hex
    table_schema = {"t": "text", "date": "timestamptz", "i": "int"}

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema=table_schema,
        pre_delete_table=True,
        logger=logger,
    )
    insert_example_data(holo)

    res = holo.search([2, 2, 2, 2, 2], k=2, select_columns=["i"], schema_data_filters={"t": "text 1"})
    assert len(res) == 1
    assert res[0]["i"] == 1

    holo.run_sql(f"drop table {table_name}", False)


def test_delete():
    table_name = "test_vector_table_" + uuid.uuid1().hex
    table_schema = {"t": "text", "date": "timestamptz", "i": "int"}

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema=table_schema,
        pre_delete_table=True,
        logger=logger,
    )

    insert_example_data(holo, use_metadata=True)
    res = holo.run_sql(f"select count(1), sum(i) from {table_name}")
    assert res[0][0] == 3
    assert res[0][1] == 3
    assert holo.size() == 3

    holo.delete_vectors(schema_data_filters={"t": "text 0"})
    res = holo.run_sql(f"select count(1), sum(i) from {table_name}")
    assert res[0][0] == 2
    assert res[0][1] == 3
    assert holo.size() == 2

    holo.delete_vectors(metadata_filters={"t": "text 1"})
    res = holo.run_sql(f"select count(1), sum(i) from {table_name}")
    assert res[0][0] == 1
    assert res[0][1] == 2
    assert holo.size() == 1

    holo.delete_vectors()
    res = holo.run_sql(f"select count(1) from {table_name}")
    assert res[0][0] == 0
    assert holo.size() == 0

    holo.run_sql(f"drop table {table_name}", False)


@pytest.mark.parametrize("distance_method", ["SquaredEuclidean", "Euclidean", "InnerProduct"])
def test_distance_method(distance_method):
    table_name = "test_vector_table_" + uuid.uuid1().hex

    holo = HologresVector(
        get_connection_string(),
        5,
        table_name=table_name,
        table_schema={"t": "text", "date": "timestamptz", "i": "int"},
        distance_method=distance_method,
        pre_delete_table=True,
        logger=logger,
    )
    insert_example_data(holo)

    res = holo.search([2, 2, 2, 2, 2], k=2, select_columns=["t", "date", "i"])
    assert len(res) == 2
    for i in range(2):
        assert res[i]["i"] == i
        assert res[i]["t"] == f"text {i}"

    holo.run_sql(f"drop table {table_name}", False)



from __future__ import annotations

import json
import logging
import uuid
from typing import Any

from hologres_vector.hologres_wrapper import HologresWrapper


class HologresVector:
    """VectorStore implementation using Hologres.

    - `connection_string` is a hologres connection string. Call `connection_string_from_db_params` to generate.
    - `ndims` is the number of dimensions of the vector
    - `table_name` is the name of the table to store vector and json data.
        - NOTE: The table will be created when initializing the store (if not exists)
            So, make sure the user has the right permissions to create tables.
    - `table_schema` defines the types of the other columns of the table.
        - Useful for storing original texts other other information related to a vector.
    - `pre_delete_table` if True, will delete the table if it exists.
        (default: False)
        - Useful for testing.
    """

    def __init__(
        self,
        connection_string: str,
        ndims: int,
        table_name: str,
        table_schema: dict[str, str] | None = None,
        distance_method: str = "SquaredEuclidean",
        pre_delete_table: bool = False,
        logger: logging.Logger | None = None,
    ) -> None:
        self.connection_string = connection_string
        self.ndims = ndims
        self.table_name = table_name
        self.table_schema = table_schema
        self.distance_method = distance_method
        self.pre_delete_table = pre_delete_table
        self.logger = logger or logging.getLogger(__name__)
        self.__post_init__()

    def __post_init__(
        self,
    ) -> None:
        """
        Initialize the store.
        """
        self.storage = HologresWrapper(
            self.connection_string, self.ndims, self.table_name, self.distance_method, self.logger
        )
        self.__create_vector_extension()
        self.storage.create_table(self.table_schema, self.pre_delete_table)

    def __create_vector_extension(self) -> None:
        try:
            self.storage.create_vector_extension()
        except Exception as e:
            self.logger.exception(e)
            raise e

    def upsert_vectors(
        self,
        vectors: list[list[float]],
        ids: list[str] | None = None,
        metadatas: list[dict] | None = None,
        schema_datas: list[dict[str, str]] | None = None,
    ) -> None:
        """Add vectors to the vectorstore.

        Args:
            vectors: List of vectors.
            metadatas: List of metadatas associated with the vectors.
        """
        if ids is None:
            ids = [str(uuid.uuid1()) for _ in vectors]
        if metadatas is None:
            metadatas = [{} for _ in vectors]
        if schema_datas is None:
            schema_datas = [{} for _ in vectors]
        try:
            self.storage.batch_upsert(ids, vectors, metadatas, schema_datas)
        except Exception as e:
            self.logger.exception(e)
            self.storage.conn.commit()
            raise e

    def search(
        self,
        vector: list[float],
        k: int = 3,
        select_columns: list[str] | None = None,
        metadata_filters: dict | None = None,
        schema_data_filters: dict | None = None,
    ) -> list[dict[str, Any]]:
        """
        Return: list[{id, metadata::text, distance, select_column_0, ... , select_column_n}]
        """
        results: list[tuple] = self.storage.query_nearest_neighbours(
            vector, k, select_columns, metadata_filters, schema_data_filters
        )
        ret = []
        for row in results:
            parsed_row = {
                "id": row[0],
                "metadata": json.loads(row[1]) if row[1] is not None and row[1] != "" else None,
                "distance": row[2],
            }
            if select_columns:
                i = 3
                for col_name in select_columns:
                    parsed_row[col_name] = row[i]
                    i += 1
                if i != len(row):
                    raise AssertionError()
            ret.append(parsed_row)

        return ret

    def query(
        self,
        select_columns: list[str] | None = None,
        metadata_filters: dict | None = None,
        schema_data_filters: dict | None = None,
        limit: int = 100,
    ) -> list[dict[str, Any]]:
        """
        Get data by filter, without using vectors. By default, limit the number of output rows to 100.
        Return: list[{id, vector, metadata::text, select_column_0, ... , select_column_n}]
        """
        results = self.storage.query_by_filters(select_columns, metadata_filters, schema_data_filters, limit)
        ret = []
        for row in results:
            parsed_row = {
                "id": row[0],
                "vector": row[1],
                "metadata": json.loads(row[2]),
            }
            if select_columns:
                i = 3
                for col_name in select_columns:
                    parsed_row[col_name] = row[i]
                    i += 1
                if i != len(row):
                    raise AssertionError()
            ret.append(parsed_row)

        return ret

    def delete_vectors(
        self,
        metadata_filters: dict | None = None,
        schema_data_filters: dict | None = None,
    ):
        """Delete by filters."""
        return self.storage.delete_by_filters(
            metadata_filters=metadata_filters, schema_data_filters=schema_data_filters
        )

    def run_sql(self, sql, fetch_results=True):
        """Use this function to run complicated queries that the SDK currently doesn't support."""
        return self.storage.run_sql(sql, fetch_results=fetch_results)

    def size(self) -> int:
        """Returns the number of rows in this collection. NOTE: This function is costly. DO NOT CALL IT FREQUENTLY."""
        return self.storage.size()

    @classmethod
    def from_data(
        cls,
        table_name: str,
        ndims: int,
        vectors: list[list[float]],
        table_schema: dict[str, str] | None = None,
        metadatas: list[dict] | None = None,
        ids: list[str] | None = None,
        schema_datas: dict[str, list[str]] | None = None,
        pre_delete_table: bool = False,
        **kwargs: Any,
    ) -> HologresVector:
        if ids is None:
            ids = [str(uuid.uuid1()) for _ in vectors]

        if not metadatas:
            metadatas = [{} for _ in vectors]

        if not schema_datas:
            schema_datas = {}

        connection_string = cls.get_connection_string(kwargs)

        store = cls(
            connection_string=connection_string,
            ndims=ndims,
            table_name=table_name,
            pre_delete_table=pre_delete_table,
            table_schema=table_schema,
        )

        store.add_vectors(vectors=vectors, metadatas=metadatas, ids=ids, schema_datas=schema_datas, **kwargs)

        return store

    @classmethod
    def connection_string_from_db_params(
        cls,
        host: str,
        port: int,
        database: str,
        user: str,
        password: str,
    ) -> str:
        """Return connection string from database parameters."""
        return f"dbname={database} user={user} password={password} host={host} port={port}"
